@@ -21,6 +21,7 @@
 #include "rmw/rmw.h"
 
 #include "fastdds/dds/subscriber/SampleInfo.hpp"
+#include "fastdds/dds/core/StackAllocatedSequence.hpp"
 
 #include "fastrtps/utils/collections/ResourceLimitedVector.hpp"
 
@@ -34,11 +35,15 @@
 #include "rmw_fastrtps_shared_cpp/TypeSupport.hpp"
 #include "rmw_fastrtps_shared_cpp/utils.hpp"
 
+#include "rosidl_dynamic_typesupport/types.h"
+
+#include "tracetools/tracetools.h"
+
+#include "rcpputils/scope_exit.hpp"
+
 namespace rmw_fastrtps_shared_cpp
 {
 
-using DataSharingKind = eprosima::fastdds::dds::DataSharingKind;
-
 void
 _assign_message_info(
   const char * identifier,
@@ -47,6 +52,11 @@ _assign_message_info(
 {
   message_info->source_timestamp = sinfo->source_timestamp.to_ns();
   message_info->received_timestamp = sinfo->reception_timestamp.to_ns();
+  auto fastdds_sn = sinfo->sample_identity.sequence_number();
+  message_info->publication_sequence_number =
+    (static_cast<uint64_t>(fastdds_sn.high) << 32) |
+    static_cast<uint64_t>(fastdds_sn.low);
+  message_info->reception_sequence_number = RMW_MESSAGE_INFO_SEQUENCE_NUMBER_UNSUPPORTED;
   rmw_gid_t * sender_gid = &message_info->publisher_gid;
   sender_gid->implementation_identifier = identifier;
   memset(sender_gid->data, 0, RMW_GID_STORAGE_SIZE);
@@ -71,44 +81,56 @@ _take(
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
     subscription handle,
     subscription->implementation_identifier, identifier,
-    return RMW_RET_INCORRECT_RMW_IMPLEMENTATION)
+    return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
   auto info = static_cast<CustomSubscriberInfo *>(subscription->data);
   RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "custom subscriber info is null", return RMW_RET_ERROR);
 
-  eprosima::fastdds::dds::SampleInfo sinfo;
-
   rmw_fastrtps_shared_cpp::SerializedData data;
-
-  data.is_cdr_buffer = false;
+  data.type = FASTRTPS_SERIALIZED_DATA_TYPE_ROS_MESSAGE;
   data.data = ros_message;
   data.impl = info->type_support_impl_;
 
-  while (0 < info->data_reader_->get_unread_count()) {
-    if (info->data_reader_->take_next_sample(&data, &sinfo) == ReturnCode_t::RETCODE_OK) {
-      // Update hasData from listener
-      info->listener_->update_has_data(info->data_reader_);
+  eprosima::fastdds::dds::StackAllocatedSequence<void *, 1> data_values;
+  const_cast<void **>(data_values.buffer())[0] = &data;
+  eprosima::fastdds::dds::SampleInfoSeq info_seq{1};
 
-      if (subscription->options.ignore_local_publications) {
-        auto sample_writer_guid =
-          eprosima::fastrtps::rtps::iHandle2GUID(sinfo.publication_handle);
+  while (ReturnCode_t::RETCODE_OK == info->data_reader_->take(data_values, info_seq, 1)) {
+    // The info->data_reader_->take() call already modified the ros_message arg
+    // See rmw_fastrtps_shared_cpp/src/TypeSupport_impl.cpp
 
-        if (sample_writer_guid.guidPrefix == info->data_reader_->guid().guidPrefix) {
-          // This is a local publication. Ignore it
-          continue;
-        }
+    auto reset = rcpputils::make_scope_exit(
+      [&]()
+      {
+        data_values.length(0);
+        info_seq.length(0);
+      });
+
+    if (subscription->options.ignore_local_publications) {
+      auto sample_writer_guid =
+        eprosima::fastrtps::rtps::iHandle2GUID(info_seq[0].publication_handle);
+
+      if (sample_writer_guid.guidPrefix == info->data_reader_->guid().guidPrefix) {
+        // This is a local publication. Ignore it
+        continue;
       }
+    }
 
-      if (sinfo.valid_data) {
-        if (message_info) {
-          _assign_message_info(identifier, message_info, &sinfo);
-        }
-        *taken = true;
-        break;
+    if (info_seq[0].valid_data) {
+      if (message_info) {
+        _assign_message_info(identifier, message_info, &info_seq[0]);
       }
+      *taken = true;
+      break;
     }
   }
 
+  TRACETOOLS_TRACEPOINT(
+    rmw_take,
+    static_cast<const void *>(subscription),
+    static_cast<const void *>(ros_message),
+    (message_info ? message_info->source_timestamp : 0LL),
+    *taken);
   return RMW_RET_OK;
 }
 
@@ -134,14 +156,6 @@ _take_sequence(
   auto info = static_cast<CustomSubscriberInfo *>(subscription->data);
   RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "custom subscriber info is null", return RMW_RET_ERROR);
 
-  // Limit the upper bound of reads to the number unread at the beginning.
-  // This prevents any samples that are added after the beginning of the
-  // _take_sequence call from being read.
-  auto unread_count = info->data_reader_->get_unread_count();
-  if (unread_count < count) {
-    count = unread_count;
-  }
-
   for (size_t ii = 0; ii < count; ++ii) {
     taken_flag = false;
     ret = _take(
@@ -183,7 +197,7 @@ __rmw_take_event(
     return RMW_RET_ERROR);
 
   auto event = static_cast<CustomEventInfo *>(event_handle->data);
-  if (event->getListener()->takeNextEvent(event_handle->event_type, event_info)) {
+  if (event->get_listener()->take_event(event_handle->event_type, event_info)) {
     *taken = true;
     return RMW_RET_OK;
   }
@@ -292,38 +306,46 @@ _take_serialized_message(
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
     subscription handle,
     subscription->implementation_identifier, identifier,
-    return RMW_RET_INCORRECT_RMW_IMPLEMENTATION)
+    return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
   auto info = static_cast<CustomSubscriberInfo *>(subscription->data);
   RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "custom subscriber info is null", return RMW_RET_ERROR);
 
   eprosima::fastcdr::FastBuffer buffer;
-  eprosima::fastdds::dds::SampleInfo sinfo;
 
   rmw_fastrtps_shared_cpp::SerializedData data;
-  data.is_cdr_buffer = true;
+  data.type = FASTRTPS_SERIALIZED_DATA_TYPE_CDR_BUFFER;
   data.data = &buffer;
-  data.impl = nullptr;    // not used when is_cdr_buffer is true
+  data.impl = nullptr;  // not used when type is FASTRTPS_SERIALIZED_DATA_TYPE_CDR_BUFFER
 
-  if (info->data_reader_->take_next_sample(&data, &sinfo) == ReturnCode_t::RETCODE_OK) {
-    // Update hasData from listener
-    info->listener_->update_has_data(info->data_reader_);
+  eprosima::fastdds::dds::StackAllocatedSequence<void *, 1> data_values;
+  const_cast<void **>(data_values.buffer())[0] = &data;
+  eprosima::fastdds::dds::SampleInfoSeq info_seq{1};
 
-    if (sinfo.valid_data) {
+  while (ReturnCode_t::RETCODE_OK == info->data_reader_->take(data_values, info_seq, 1)) {
+    auto reset = rcpputils::make_scope_exit(
+      [&]()
+      {
+        data_values.length(0);
+        info_seq.length(0);
+      });
+
+    if (info_seq[0].valid_data) {
       auto buffer_size = static_cast<size_t>(buffer.getBufferSize());
       if (serialized_message->buffer_capacity < buffer_size) {
         auto ret = rmw_serialized_message_resize(serialized_message, buffer_size);
         if (ret != RMW_RET_OK) {
-          return ret;  // Error message already set
+          return ret;           // Error message already set
         }
       }
       serialized_message->buffer_length = buffer_size;
       memcpy(serialized_message->buffer, buffer.getBuffer(), serialized_message->buffer_length);
 
       if (message_info) {
-        _assign_message_info(identifier, message_info, &sinfo);
+        _assign_message_info(identifier, message_info, &info_seq[0]);
       }
       *taken = true;
+      break;
     }
   }
 
@@ -376,13 +398,114 @@ __rmw_take_serialized_message_with_info(
     identifier, subscription, serialized_message, taken, message_info, allocation);
 }
 
+rmw_ret_t
+_take_dynamic_message(
+  const char * identifier,
+  const rmw_subscription_t * subscription,
+  rosidl_dynamic_typesupport_dynamic_data_t * dynamic_data,
+  bool * taken,
+  rmw_message_info_t * message_info,
+  rmw_subscription_allocation_t * allocation)
+{
+  (void) allocation;
+  *taken = false;
+
+  RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
+    subscription handle,
+    subscription->implementation_identifier, identifier,
+    return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
+
+  auto info = static_cast<CustomSubscriberInfo *>(subscription->data);
+  RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "custom subscriber info is null", return RMW_RET_ERROR);
+
+  eprosima::fastcdr::FastBuffer buffer;
+
+  rmw_fastrtps_shared_cpp::SerializedData data;
+  data.type = FASTRTPS_SERIALIZED_DATA_TYPE_DYNAMIC_MESSAGE;
+  data.data = dynamic_data->impl.handle;
+  data.impl = nullptr;  // not used when type is FASTRTPS_SERIALIZED_DATA_TYPE_DYNAMIC_MESSAGE
+
+  eprosima::fastdds::dds::StackAllocatedSequence<void *, 1> data_values;
+  const_cast<void **>(data_values.buffer())[0] = &data;
+  eprosima::fastdds::dds::SampleInfoSeq info_seq{1};
+
+  while (ReturnCode_t::RETCODE_OK == info->data_reader_->take(data_values, info_seq, 1)) {
+    // The info->data_reader_->take() call already modified the dynamic_data arg
+    // See rmw_fastrtps_shared_cpp/src/TypeSupport_impl.cpp
+
+    auto reset = rcpputils::make_scope_exit(
+      [&]()
+      {
+        data_values.length(0);
+        info_seq.length(0);
+      });
+
+    if (info_seq[0].valid_data) {
+      if (message_info) {
+        _assign_message_info(identifier, message_info, &info_seq[0]);
+      }
+      *taken = true;
+      break;
+    }
+  }
+
+  return RMW_RET_OK;
+}
+
+rmw_ret_t
+__rmw_take_dynamic_message(
+  const char * identifier,
+  const rmw_subscription_t * subscription,
+  rosidl_dynamic_typesupport_dynamic_data_t * dynamic_data,
+  bool * taken,
+  rmw_subscription_allocation_t * allocation)
+{
+  RMW_CHECK_ARGUMENT_FOR_NULL(
+    subscription, RMW_RET_INVALID_ARGUMENT);
+
+  RMW_CHECK_ARGUMENT_FOR_NULL(
+    dynamic_data, RMW_RET_INVALID_ARGUMENT);
+
+  RMW_CHECK_ARGUMENT_FOR_NULL(
+    taken, RMW_RET_INVALID_ARGUMENT);
+
+  return _take_dynamic_message(
+    identifier, subscription, dynamic_data, taken, nullptr, allocation);
+}
+
+rmw_ret_t
+__rmw_take_dynamic_message_with_info(
+  const char * identifier,
+  const rmw_subscription_t * subscription,
+  rosidl_dynamic_typesupport_dynamic_data_t * dynamic_data,
+  bool * taken,
+  rmw_message_info_t * message_info,
+  rmw_subscription_allocation_t * allocation)
+{
+  RMW_CHECK_ARGUMENT_FOR_NULL(
+    subscription, RMW_RET_INVALID_ARGUMENT);
+
+  RMW_CHECK_ARGUMENT_FOR_NULL(
+    dynamic_data, RMW_RET_INVALID_ARGUMENT);
+
+  RMW_CHECK_ARGUMENT_FOR_NULL(
+    taken, RMW_RET_INVALID_ARGUMENT);
+
+  RMW_CHECK_ARGUMENT_FOR_NULL(
+    message_info, RMW_RET_INVALID_ARGUMENT);
+
+  return _take_dynamic_message(
+    identifier, subscription, dynamic_data, taken, message_info, allocation);
+}
+
 // ----------------- Loans related code ------------------------- //
 
 struct GenericSequence : public eprosima::fastdds::dds::LoanableCollection
 {
   GenericSequence() = default;
 
-  void resize(size_type /*new_length*/) override
+  void resize(
+    size_type /*new_length*/) override
   {
     // This kind of collection should only be used with loans
     throw std::bad_alloc();
@@ -397,18 +520,21 @@ struct LoanManager
     eprosima::fastdds::dds::SampleInfoSeq info_seq{};
   };
 
-  explicit LoanManager(const eprosima::fastrtps::ResourceLimitedContainerConfig & items_cfg)
+  explicit LoanManager(
+    const eprosima::fastrtps::ResourceLimitedContainerConfig & items_cfg)
   : items(items_cfg)
   {
   }
 
-  void add_item(std::unique_ptr<Item> item)
+  void add_item(
+    std::unique_ptr<Item> item)
   {
     std::lock_guard<std::mutex> guard(mtx);
     items.push_back(std::move(item));
   }
 
-  std::unique_ptr<Item> erase_item(void * loaned_message)
+  std::unique_ptr<Item> erase_item(
+    void * loaned_message)
   {
     std::unique_ptr<Item> ret{nullptr};
 
@@ -436,8 +562,7 @@ __init_subscription_for_loans(
 {
   auto info = static_cast<CustomSubscriberInfo *>(subscription->data);
   const auto & qos = info->data_reader_->get_qos();
-  bool has_data_sharing = DataSharingKind::OFF != qos.data_sharing().kind();
-  subscription->can_loan_messages = has_data_sharing && info->type_support_->is_plain();
+  subscription->can_loan_messages = info->type_support_->is_plain();
   if (subscription->can_loan_messages) {
     const auto & allocation_qos = qos.reader_resource_limits().outstanding_reads_allocation;
     info->loan_manager_ = std::make_shared<LoanManager>(allocation_qos);
@@ -475,7 +600,6 @@ __rmw_take_loaned_message_internal(
       }
       *loaned_message = item->data_seq.buffer()[0];
       *taken = true;
-      info->listener_->update_has_data(info->data_reader_);
 
       info->loan_manager_->add_item(std::move(item));
 
@@ -488,7 +612,6 @@ __rmw_take_loaned_message_internal(
 
   // No data available, return loan information.
   *taken = false;
-  info->listener_->update_has_data(info->data_reader_);
   return RMW_RET_OK;
 }
 
@@ -523,4 +646,5 @@ __rmw_return_loaned_message_from_subscription(
   RMW_SET_ERROR_MSG("Trying to return message not loaned by this subscription");
   return RMW_RET_ERROR;
 }
+
 }  // namespace rmw_fastrtps_shared_cpp
